# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
"""Configuration subsystem."""

import configparser
import os.path
import re
from configparser import ConfigParser, NoOptionError
from enum import Enum
from typing import Dict, Optional, TypedDict, Union

from xdg import XDG_CONFIG_HOME

# CONSTANTS
CONFIG = "config"
REPLICATION = "replication"
DEFAULT_CONFIG = CONFIG


# Type Definitions
class Remote(TypedDict):
    """Remote Dictionary definition."""

    owner: str
    remotenamestyle: str
    repo: str


PlatformType = Dict[str, Remote]


class ReplicationRemotes(TypedDict, total=False):
    """Platform Dictionary definition."""

    github: PlatformType
    gitlab: PlatformType


class Platform(Enum):
    """Enumeration of all platforms recognized by the app."""

    GITHUB = "github"
    GITLAB = "gitlab"


G2P_CONFIG_FILE = os.path.join(
    XDG_CONFIG_HOME, "gerrit_to_platform", "gerrit_to_platform.ini"
)

GERRIT_REPLICATION_CONFIG = os.path.join(
    XDG_CONFIG_HOME, "gerrit_to_platform", "replication.config"
)

CONFIG_FILES = {
    CONFIG: G2P_CONFIG_FILE,
    REPLICATION: GERRIT_REPLICATION_CONFIG,
}


def get_config(config_type: str = DEFAULT_CONFIG) -> ConfigParser:
    """
    Get the config object.

    Args:
        config_type (str): Type of configuration file that the parser should use

    Returns:
        ConfigParser: A loaded ConfigParser object with the requested
            configuration file
    """
    config = configparser.ConfigParser()
    conf_file = CONFIG_FILES[config_type]
    with open(conf_file) as config_file:
        config.read_file(config_file, conf_file)
    return config


def get_mapping(mapping_section: str) -> Union[Dict[str, str], None]:
    """
    Return all of the keyword to job mappings

    Args:
        mapping_section (str): the section of the config file to load

    Returns:
        Optional(Dict[str,str]): The key / value mapping object or None if it
            does not exist
    """
    config = get_config()

    section = f'mapping "{mapping_section}"'
    if section in config.sections():
        return dict(config.items(section))

    return None


def get_replication_remotes() -> ReplicationRemotes:
    """
    Get the replication remotes available.

    Returns:
        ReplicationRemotes: All the replication remotes defined the Gerrit
            configuration file
    """
    remotes_config = get_config(REPLICATION)
    remotes: ReplicationRemotes = {}

    for section in remotes_config.sections():
        if 'remote "' not in section:
            continue

        subsect = r'"(.*)"'
        subsection = re.findall(subsect, section)[0]
        url = remotes_config.get(section, "url")
        try:
            authgroup = remotes_config.get(section, "authgroup")
        except NoOptionError:
            authgroup = ""
        try:
            namestyle = remotes_config.get(section, "remotenamestyle")
        except NoOptionError:
            namestyle = "slash"

        if (
            "github" in section.lower()
            or "github" in subsection.lower()
            or "github" in authgroup.lower()
        ):
            platform_type = Platform.GITHUB.value
        elif (
            "gitlab" in section.lower()
            or "gitlab" in subsection.lower()
            or "gitlab" in authgroup.lower()
        ):
            platform_type = Platform.GITLAB.value
        else:
            continue

        url_regex = r":(.*)/(.*)$"
        owner, repo = re.findall(url_regex, url)[0]

        if platform_type in remotes.keys():
            remotes[platform_type][subsection] = {  # type: ignore
                "owner": owner,
                "remotenamestyle": namestyle,
                "repo": repo,
            }
        else:
            remotes[platform_type] = {  # type: ignore
                subsection: {
                    "owner": owner,
                    "remotenamestyle": namestyle,
                    "repo": repo,
                }
            }

    return remotes


def has_section(section: str) -> bool:
    """
    Indicate if section exists in config file.

    Args:
        section (str): config section to lookup

    Returns:
        bool: True if the section exists in configuration, False otherwise
    """
    config = get_config()
    return config.has_section(section)


def get_setting(section: str, option: Optional[str] = None) -> Union[list, str]:
    """
    Get all configuration options from a section, or specific option from a
    section.

    Args:
        section (str): config section to use
        option (Optional[str]): a suboption to get from the section

    Returns:
        list: all of the configuration for a section
        str: the single sub-option of a section
    """
    config = get_config()

    if option:
        return config.get(section, option)

    return config.options(section)



# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
"""Unit tests for github."""

import json
import os
from typing import Any, Callable, Dict, List, Union

import gerrit_to_platform.github as github  # type: ignore
import gerrit_to_platform.helpers  # type: ignore
from gerrit_to_platform.config import Platform  # type: ignore
from gerrit_to_platform.helpers import (  # type: ignore
    choose_dispatch,
    choose_filter_workflows,
    convert_repo_name,
    find_and_dispatch,
    get_change_id,
    get_change_number,
    get_change_refspec,
    get_magic_repo,
)

FIXTURE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "fixtures",
)

VERIFY_FILTERED_WORKFLOWS = os.path.join(
    FIXTURE_DIR, "github_workflow_list_filter_workflows_return.json"
)

MERGE_FILTERED_WORKFLOWS = os.path.join(
    FIXTURE_DIR, "github_workflow_list_filter_workflows_return_merge.json"
)

REPLICATION_REMOTES = os.path.join(FIXTURE_DIR, "replication_remotes_return.json")

REPLICATION_REMOTES_GITHUB = os.path.join(
    FIXTURE_DIR, "limited_replication_remotes_return_github.json"
)

REPLICATION_REMOTES_GITLAB = os.path.join(
    FIXTURE_DIR, "limited_replication_remotes_return_gitlab.json"
)

REQUIRED_VERIFY_FILTERED_WORKFLOWS = os.path.join(
    FIXTURE_DIR, "github_workflow_list_required_filter_workflows_return.json"
)

PATCH1_GERRIT_VERIFY = (
    "Dispatching workflow 'Gerrit Verify', id 20937807 on "
    + "github:example/example-project for change 1 patch 1"
)
PATCH1_VERIFY = (
    "Dispatching workflow 'Verify', id 18525370 on "
    + "github:example/example-project for change 1 patch 1"
)
PATCH1_REQUIRED_VERIFY = (
    "Dispatching required workflow 'Required Gerrit Verify', id 20937808 on "
    + "github:example/.github for change 1 patch 1 against "
    + "github:example/example-project"
)
PATCH1_CHECK_MAIN = (
    "Dispatching workflow 'Check Main', id 17098575 on "
    + "github:example/example-project for change 1 patch 1"
)
CHANGE2_MERGE = (
    "Dispatching workflow 'Gerrit Merge', id 18525370 on "
    + "github:example/example-project for change 1 patch 1"
)


def test_choose_dispatch(mocker):
    """Test choose_dispatch."""
    expected = github.dispatch_workflow
    actual = choose_dispatch(Platform.GITHUB)
    assert expected == actual

    expected = None
    actual = choose_dispatch(Platform.GITLAB)
    assert expected == actual


def test_choose_filter_workflows(mocker):
    """Test choose_filter_workflows."""
    expected = github.filter_workflows
    actual = choose_filter_workflows(Platform.GITHUB)
    assert expected == actual

    expected = None
    actual = choose_filter_workflows(Platform.GITLAB)
    assert expected == actual


def test_convert_repo_name(mocker):
    """Convert repository name to platform style."""

    with open(REPLICATION_REMOTES) as remotes:
        replication_remotes = json.load(remotes)

    expected = "foo-bar"
    actual = convert_repo_name(
        replication_remotes, Platform.GITHUB, "github", "foo/bar"
    )
    assert expected == actual
    expected = "foo_bar"
    actual = convert_repo_name(
        replication_remotes, Platform.GITHUB, "github-other", "foo/bar"
    )
    assert expected == actual
    expected = "foo/bar"
    actual = convert_repo_name(
        replication_remotes, Platform.GITLAB, "gitlab", "foo/bar"
    )
    assert expected == actual


def test_find_and_dispatch(mocker, capfd):
    """Test the find_and_dispatch helper."""
    inputs = {
        "GERRIT_BRANCH": "main",
        "GERRIT_CHANGE_ID": "Ichange_id",
        "GERRIT_CHANGE_NUMBER": "1",
        "GERRIT_CHANGE_URL": "https://foo.bar/r/c/example/+/1",
        "GERRIT_EVENT_TYPE": "test",
        "GERRIT_PATCHSET_NUMBER": "1",
        "GERRIT_PATCHSET_REVISION": "foo",
        "GERRIT_PROJECT": "example/project",
        "GERRIT_REFSPEC": "refs/heads/main",
    }

    with open(REPLICATION_REMOTES_GITHUB) as remotes:
        replication_remotes = json.load(remotes)

    mocker.patch(
        "gerrit_to_platform.helpers.get_replication_remotes",
        return_value=replication_remotes,
    )

    def mock_filter_workflows(
        owner: str, repo: str, search_filter: str, search_required: bool = False
    ) -> List[Dict[str, str]]:
        """Mock of filter_workflows."""
        filter_file = VERIFY_FILTERED_WORKFLOWS

        if search_required:
            filter_file = REQUIRED_VERIFY_FILTERED_WORKFLOWS

        if search_filter == "merge":
            filter_file = MERGE_FILTERED_WORKFLOWS

        with open(filter_file) as workflows:
            return json.load(workflows)

    def mock_choose_filter_workflows(platform: Platform) -> Union[Callable, None]:
        """Mock of choose_filter_workflows."""
        if platform == Platform.GITHUB:
            return mock_filter_workflows

        return None

    mocker.patch.object(
        gerrit_to_platform.helpers,
        "choose_filter_workflows",
        mock_choose_filter_workflows,
    )

    def mock_dispatch_workflow(
        owner: str, repository: str, workflow_id: str, ref: str, inputs: Dict[str, str]
    ) -> Any:
        """Mock of dispach_workflow"""
        return {}

    def mock_choose_dispatch(platform: Platform) -> Union[Callable, None]:
        """Mock of choose_dispacth."""
        if platform == Platform.GITHUB:
            return mock_dispatch_workflow

        return None

    mocker.patch.object(
        gerrit_to_platform.helpers,
        "choose_dispatch",
        mock_choose_dispatch,
    )

    find_and_dispatch("example-project", "verify", inputs)
    actual = capfd.readouterr().out
    assert PATCH1_GERRIT_VERIFY in actual
    assert PATCH1_VERIFY not in actual
    assert PATCH1_REQUIRED_VERIFY in actual
    assert PATCH1_CHECK_MAIN not in actual

    find_and_dispatch("example-project", "merge", inputs)
    actual = capfd.readouterr().out
    assert CHANGE2_MERGE in actual

    with open(REPLICATION_REMOTES_GITLAB) as remotes:
        replication_remotes = json.load(remotes)
    mocker.patch(
        "gerrit_to_platform.helpers.get_replication_remotes",
        return_value=replication_remotes,
    )
    find_and_dispatch("example-project", "verify", inputs)
    actual = capfd.readouterr().out
    assert PATCH1_GERRIT_VERIFY not in actual
    assert PATCH1_VERIFY not in actual
    assert PATCH1_REQUIRED_VERIFY not in actual
    assert PATCH1_CHECK_MAIN not in actual
    assert actual == ""


def test_get_change_id(mocker):
    """Test get_change_id"""
    expected = "Ibaz"
    actual = get_change_id("foo~bar~Ibaz")
    assert expected == actual


def test_get_change_number(mocker):
    """Test get_change_number"""
    expected = "71001"
    actual = get_change_number("https://example.org.org/r/c/example/+/71001")
    assert expected == actual


def test_get_change_refspec(mocker):
    """Test get_change_refspec"""
    expected = "refs/changes/01/1/1"
    actual = get_change_refspec("1", "1")
    assert expected == actual

    expected = "refs/changes/01/1/2"
    actual = get_change_refspec("1", "2")
    assert expected == actual

    expected = "refs/changes/02/202/1"
    actual = get_change_refspec("202", "1")
    assert expected == actual


def test_get_magic_repo(mocker):
    """Test get_magic_repo"""
    expected = ".github"
    actual = get_magic_repo(Platform.GITHUB)
    assert expected == actual

    expected = None
    actual = get_magic_repo(Platform.GITLAB)
    assert expected == actual



# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
"""
    Dummy conftest.py for gerrit_to_platform.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""

# import pytest



# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
"""Unit tests for comment_added."""

from typing import Dict

from typer.testing import CliRunner

import gerrit_to_platform.comment_added  # type: ignore
from gerrit_to_platform.comment_added import app  # type: ignore

CHANGE1 = [
    "--change=example/project~master~I308b4eda73ff90ee486f14e01db145684889eaae",
    "--change-url=https://gerrit.example.org/r/c/example/project/+/1",
    "--change-owner='Foo <foo@example.org>'",
    "--change-owner-username=foo",
    "--project=example/project",
    "--branch=master",
    "--topic=testing",
    "--author='Foo <foo@example.org>'",
    "--author-username=foo",
    "--commit=7f0f8a2b05546d5956b0bb1431ba13c8cbe94631",
    """--comment=Patch Set 1:

recheck""",
    "--Code-Review=0",
]

CHANGE2 = [
    "--change=example/project~master~I308b4eda73ff90ee486f14e01db145684889eaae",
    "--change-url=https://gerrit.example.org/r/c/example/project/+/512",
    "--change-owner='Foo <foo@example.org>'",
    "--change-owner-username=foo",
    "--project=example/project",
    "--branch=master",
    "--topic=testing",
    "--author='Foo <foo@example.org>'",
    "--author-username=foo",
    "--commit=7f0f8a2b05546d5956b0bb1431ba13c8cbe94631",
    """--comment=Patch Set 6:

remerge""",
    "--Code-Review=0",
]

CHANGE1_VERIFY = "Dispatched verify for patchset 1"
CHANGE2_MERGE = "Dispatched merge for patchset 6"

runner = CliRunner()


def test_app(mocker):
    """Test app options exist."""
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0

    mocker.patch(
        "gerrit_to_platform.comment_added.get_mapping",
        return_value={"recheck": "verify", "remerge": "merge"},
    )

    def mock_find_and_dispatch(
        project: str, workflow_filter: str, inputs: Dict[str, str]
    ):
        """Mock find_and_dispatch helper."""
        print(
            f"Dispatched {workflow_filter} for patchset {inputs['GERRIT_PATCHSET_NUMBER']}"
        )

    mocker.patch.object(
        gerrit_to_platform.comment_added,
        "find_and_dispatch",
        mock_find_and_dispatch,
    )

    result = runner.invoke(app, CHANGE1)
    assert result.exit_code == 0
    assert CHANGE1_VERIFY in result.stdout
    assert CHANGE2_MERGE not in result.stdout

    result = runner.invoke(app, CHANGE2)
    assert result.exit_code == 0
    assert CHANGE1_VERIFY not in result.stdout
    assert CHANGE2_MERGE in result.stdout

    mocker.patch("gerrit_to_platform.comment_added.get_mapping", return_value=None)
    result = runner.invoke(app, CHANGE1)
    assert result.exit_code == 0
    assert "" == result.stdout



# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
"""Unit tests for config."""

import json
import os

import pytest

import gerrit_to_platform.config  # type: ignore
from gerrit_to_platform.config import (  # type: ignore
    CONFIG,
    REPLICATION,
    get_config,
    get_mapping,
    get_replication_remotes,
    get_setting,
    has_section,
)

FIXTURE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "fixtures",
)

TEST_CONFIG = os.path.join(FIXTURE_DIR, "testconfig.ini")
REPLICATION_CONFIG = os.path.join(FIXTURE_DIR, "replication.config")

MOCK_CONFIG_FILES = {
    CONFIG: TEST_CONFIG,
    REPLICATION: REPLICATION_CONFIG,
}


def test_get_config(mocker):
    """Test getting config data."""
    mocker.patch.object(
        gerrit_to_platform.config,
        "CONFIG_FILES",
        MOCK_CONFIG_FILES,
    )
    assert get_config().has_section("github.com")
    assert get_config(REPLICATION).has_section('remote "github"')


def test_get_mapping(mocker):
    """Test get_mapping"""
    mocker.patch.object(
        gerrit_to_platform.config,
        "CONFIG_FILES",
        MOCK_CONFIG_FILES,
    )
    expected = {"recheck": "verify", "reverify": "verify", "remerge": "merge"}
    actual = get_mapping("comment-added")
    assert expected == actual

    expected = None
    actual = get_mapping("foo")
    assert expected == actual


def test_get_replication_remotes(mocker):
    """Test getting replication remotes."""
    mocker.patch.object(
        gerrit_to_platform.config,
        "CONFIG_FILES",
        MOCK_CONFIG_FILES,
    )
    REPLICATION_REMOTES_RETURN = os.path.join(
        FIXTURE_DIR, "replication_remotes_return.json"
    )
    with open(REPLICATION_REMOTES_RETURN) as remotes_return:
        expected = json.load(remotes_return)
    actual = get_replication_remotes()
    assert expected == actual


def test_has_section(mocker):
    """Test has_section function."""
    mocker.patch.object(
        gerrit_to_platform.config,
        "CONFIG_FILES",
        MOCK_CONFIG_FILES,
    )
    expected = True
    actual = has_section("github.com")
    assert expected == actual
    expected = False
    actual = has_section("foo")
    assert expected == actual


def test_get_setting(mocker):
    """Test get_setting function."""
    mocker.patch.object(
        gerrit_to_platform.config,
        "CONFIG_FILES",
        MOCK_CONFIG_FILES,
    )
    expected = ["user", "token"]
    actual = get_setting("github.com")
    assert expected == actual

    expected = "foo"
    actual = get_setting("github.com", "user")
    assert expected == actual

    with pytest.raises(Exception, match="No section: 'foobar'"):
        get_setting("foobar")
    with pytest.raises(Exception, match="No option 'baz' in section: 'github.com'"):
        get_setting("github.com", "baz")



# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
"""Handler for patchset-created events."""

import re

import typer
from typing_extensions import Annotated

from gerrit_to_platform.config import get_mapping
from gerrit_to_platform.helpers import (
    find_and_dispatch,
    get_change_id,
    get_change_number,
    get_change_refspec,
)

app = typer.Typer()


@app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def comment_added(
    context: typer.Context,
    change: Annotated[str, typer.Option(help="change id")],
    change_url: Annotated[str, typer.Option(help="change url")],
    change_owner: Annotated[str, typer.Option(help="change owner")],
    change_owner_username: Annotated[str, typer.Option(help="username")],
    project: Annotated[str, typer.Option(help="project name")],
    branch: Annotated[str, typer.Option(help="branch")],
    topic: Annotated[str, typer.Option(help="topic")],
    author: Annotated[str, typer.Option(help="comment author")],
    author_username: Annotated[str, typer.Option(help="username")],
    commit: Annotated[str, typer.Option(help="sha1")],
    comment: Annotated[str, typer.Option(help="comment")],
):
    """
    Handle comment-added hook.

    Approval scores should be added as --<approval category id> <score>

    When a score is removed it should be --<approval category id>-oldValue <score>

    Multiple scores and old scores may be passed

    ex: --Code-Review +1 --Code-Review-oldValue 0

    Args:
        context (typer.Context): handler for the typer context to allow for
            extra non-defined parameters (code scores, see above)
        change (str): change ID
        change_url (str): change URL
        change_owner (str): change owner eg: 'Foo <foo@example.com>'
        change_owner_username (str): change owner username eg: 'foo'
        project (str): Gerrit project name
        branch (str): branch change is against
        topic (str): topic change is part of
        submitter (str): submitter of change eg: 'Foo <foo@example.com>'
        submitter_username (str): submitter of change username eg: 'foo'
        commit (str): SHA1 of commit
        comment (str): the comment added to the change
    """
    change_id = get_change_id(change)
    change_number = get_change_number(change_url)

    patchset_regex = r"^Patch Set (\d+):"
    patchset = re.findall(patchset_regex, comment)[0]

    refspec = get_change_refspec(change_number, patchset)

    inputs = {
        "GERRIT_BRANCH": branch,
        "GERRIT_CHANGE_ID": change_id,
        "GERRIT_CHANGE_NUMBER": change_number,
        "GERRIT_CHANGE_URL": change_url,
        "GERRIT_EVENT_TYPE": "comment-added",
        "GERRIT_PATCHSET_NUMBER": patchset,
        "GERRIT_PATCHSET_REVISION": commit,
        "GERRIT_PROJECT": project,
        "GERRIT_REFSPEC": refspec,
    }

    mapping = get_mapping("comment-added")
    if mapping is None:
        return

    for mapper in mapping:
        if not re.findall(mapper, comment):
            continue

        find_and_dispatch(project, mapping[mapper], inputs)


if __name__ == "__main__":
    app()



# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
"""Unit tests for patchset_created."""

from typing import Dict

from typer.testing import CliRunner

import gerrit_to_platform.helpers  # type: ignore
import gerrit_to_platform.patchset_created  # type: ignore
from gerrit_to_platform.patchset_created import app  # type: ignore

CHANGE1 = [
    "--change=example/project~master~I308b4eda73ff90ee486f14e01db145684889eaae",
    "--kind=REWORK",
    "--change-url=https://gerrit.example.org/r/c/example/project/+/1",
    "--change-owner='Foo <foo@example.org>'",
    "--change-owner-username=foo",
    "--project=example/project",
    "--branch=master",
    "--topic=testing",
    "--uploader=Foo <foo@example.org>",
    "--uploader-username=foo",
    "--commit=7f0f8a2b05546d5956b0bb1431ba13c8cbe94631",
    "--patchset=1",
]


runner = CliRunner()


def test_app(mocker):
    """Test app options exist."""
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0

    def mock_find_and_dispatch(
        project: str, workflow_filter: str, inputs: Dict[str, str]
    ):
        """Mock find_and_dispatch helper."""
        print("Dispatched")

    mocker.patch.object(
        gerrit_to_platform.patchset_created,
        "find_and_dispatch",
        mock_find_and_dispatch,
    )

    result = runner.invoke(app, CHANGE1)
    assert result.exit_code == 0
    assert "Dispatched" in result.stdout



# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
"""Unit tests for github."""

import json
import os
from typing import Dict

from fastcore.net import HTTP404NotFoundError  # type: ignore

import gerrit_to_platform.config  # type: ignore
import gerrit_to_platform.github  # type: ignore
from gerrit_to_platform.config import CONFIG, REPLICATION
from gerrit_to_platform.github import (  # type: ignore
    dispatch_workflow,
    filter_path,
    filter_workflows,
    get_workflows,
)

FIXTURE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "fixtures",
)

TEST_CONFIG = os.path.join(FIXTURE_DIR, "testconfig.ini")
REPLICATION_CONFIG = os.path.join(FIXTURE_DIR, "replication.config")

MOCK_CONFIG_FILES = {
    CONFIG: TEST_CONFIG,
    REPLICATION: REPLICATION_CONFIG,
}

GITHUB_WORKFLOW_LIST = os.path.join(FIXTURE_DIR, "github_workflow_list.json")
GITHUB_WORKFLOW_LIST_RETURN = os.path.join(
    FIXTURE_DIR, "github_workflow_list_get_workflows_return.json"
)

GITHUB_EMPTY_WORKFLOW_LIST = os.path.join(
    FIXTURE_DIR, "github_workflow_empty_list.json"
)
GITHUB_EMPTY_WORKFLOW_LIST_RETURN = os.path.join(
    FIXTURE_DIR, "github_workflow_empty_list_get_workflows_return.json"
)

GITHUB_FILTERED_LIST = os.path.join(
    FIXTURE_DIR, "github_workflow_list_filter_workflows_return.json"
)

GITHUB_REQUIRED_FILTERED_LIST = os.path.join(
    FIXTURE_DIR, "github_workflow_list_required_filter_workflows_return.json"
)


def test_dispatch_workflow(mocker):
    """Test workflow triggering."""
    mocker.patch.object(
        gerrit_to_platform.config,
        "CONFIG_FILES",
        MOCK_CONFIG_FILES,
    )

    def mock_create_workflow_dispatch(
        owner: str, repository: str, workflow_id: str, ref: str, inputs: Dict[str, str]
    ):
        """Mock GhApi.actions.create_workflow_dispatch."""
        return {}

    mock_GhApi = mocker.MagicMock()
    mock_GhApi.actions.create_workflow_dispatch = mock_create_workflow_dispatch
    mocker.patch(
        "gerrit_to_platform.github.GhApi",
        return_value=mock_GhApi,
    )

    expected = {}
    actual = dispatch_workflow(
        "example_org", "example_repo", "48166297", "refs/heads/main", {}
    )
    assert expected == actual


def test_get_workflows(mocker):
    """Test workflow acquisition."""
    mocker.patch.object(
        gerrit_to_platform.config,
        "CONFIG_FILES",
        MOCK_CONFIG_FILES,
    )

    with open(GITHUB_WORKFLOW_LIST) as list_file:
        workflow_list_json = json.load(list_file)

    def mock_list_repo_workflows(owner: str, repository: str) -> dict:
        return workflow_list_json

    def mock_list_repo_workflows_exception(owner: str, repository: str) -> dict:
        raise HTTP404NotFoundError(
            mocker.MagicMock(), mocker.MagicMock(), mocker.MagicMock()
        )

    mock_GhApi = mocker.MagicMock()
    mock_GhApi.actions.list_repo_workflows = mock_list_repo_workflows

    mocker.patch(
        "gerrit_to_platform.github.GhApi",
        return_value=mock_GhApi,
    )
    with open(GITHUB_WORKFLOW_LIST_RETURN) as list_file:
        expected = json.load(list_file)
    actual = get_workflows("example", "repository")
    assert expected == actual

    with open(GITHUB_EMPTY_WORKFLOW_LIST) as list_file:
        workflow_list_json = json.load(list_file)
    with open(GITHUB_EMPTY_WORKFLOW_LIST_RETURN) as list_file:
        expected = json.load(list_file)
    actual = get_workflows("example", "repository")
    assert expected == actual

    mock_GhApi.actions.list_repo_workflows = mock_list_repo_workflows_exception
    actual = get_workflows("example", "repository")
    assert expected == actual


def test_filter_path(mocker):
    """Test filter_path"""
    # use upper case to validate that filter works case insensitive
    workflow = {"path": ".GITHUB/WORKFLOWS/GERRIT-VERIFY.yaml"}

    expected = True
    actual = filter_path("verify", workflow)
    assert actual == expected

    expected = False
    actual = filter_path("merge", workflow)
    assert actual == expected


def test_filter_workflows(mocker):
    """Return workflows that match filter."""
    with open(GITHUB_WORKFLOW_LIST_RETURN) as list_file:
        get_workflows_return = json.load(list_file)
    with open(GITHUB_FILTERED_LIST) as list_file:
        expected = json.load(list_file)

    mocker.patch(
        "gerrit_to_platform.github.get_workflows", return_value=get_workflows_return
    )

    actual = filter_workflows("example", "repository", "verify")
    assert expected == actual

    with open(GITHUB_REQUIRED_FILTERED_LIST) as list_file:
        expected = json.load(list_file)
    actual = filter_workflows("example", "repository", "verify", True)
    assert expected == actual



# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
"""Github connection module."""

from typing import Any, Dict, List

from fastcore.net import HTTP404NotFoundError  # type: ignore
from ghapi.all import GhApi  # type: ignore

from gerrit_to_platform.config import get_setting


def dispatch_workflow(
    owner: str, repository: str, workflow_id: str, ref: str, inputs: Dict[str, str]
) -> Any:
    """
    Trigger target workflow on GitHub.

    Args:
        owner (str): GitHub owner (user or organization)
        repository (str): target repository
        workflow_id (str): ID of the workflow to trigger
        ref (str): the commit ref to trigger with
        inputs (Dict[str, str]): dictionary of key / value pairs to pass as
            inputs to the workflow
    """
    github_token = get_setting("github.com", "token")
    api = GhApi(token=github_token)

    return api.actions.create_workflow_dispatch(
        owner, repository, workflow_id, ref, inputs
    )


def filter_path(search_filter: str, workflow: Dict[str, str]) -> bool:
    """
    Case insensitive path filter for use in lambda filters.

    Args:
        search_filter (str): string to search workflow file names for
        workflow (Dict[str, str]): dictionary containing all data related to
            the workflow being evaluated

    Returns:
        bool: True if search_filter matches in workflow["path"], False otherwise
    """
    path = workflow["path"].lower()
    if path.find(search_filter.lower()) >= 0:
        return True

    return False


def filter_workflows(
    owner: str, repository: str, search_filter: str, search_required: bool = False
) -> List[Dict[str, str]]:
    """
    Return a case insensitive filtered list of workflows.

    All workflows must meet the basic filtering of containing the search_filter
    in the name as well as "gerrit" in the name.

    If search_required is true, then the list will contiain all workflows that
    contain "required" in the name

    If search_required is false (the default), then the list will _exclude_ all
    workflows that contain "required" in the name.

    Args:
        owner (str): GitHub owner (entity or organization)
        repository (str): target repository
        search_filter (str): the substring to search for in workflow filenames
        search_required (bool): if workflows with "required" in the filename
            are to be returned. If false, then required workflows will be
            filtered out, if true only required workflows will be returned.

    Returns:
        List[Dict[str, str]]: list of dictionaries containing all workflows
            that meet the search criteria of having "gerrit" in the name, the
            search_filter substring, and either required or not required
            according to the search_required argument.
    """
    workflows = get_workflows(owner, repository)
    filtered_workflows: List[Dict[str, str]] = []

    filtered_workflows = list(
        filter(lambda workflow: filter_path(search_filter, workflow), workflows)
    )
    filtered_workflows = list(
        filter(lambda workflow: filter_path("gerrit", workflow), filtered_workflows)
    )
    if search_required:
        filtered_workflows = list(
            filter(
                lambda workflow: filter_path("required", workflow), filtered_workflows
            )
        )
    else:
        filtered_workflows = list(
            filter(
                lambda workflow: not filter_path("required", workflow),
                filtered_workflows,
            )
        )

    return filtered_workflows


def get_workflows(owner: str, repository: str) -> List[Dict[str, str]]:
    """
    Get all active workflows for specific repository.

    Args:
        owner (str): GitHub owner (entity or organization)
        repository (str): target repository

    Returns:
        List[Dict[str, str]]: list of dictionaries containing data related to
            active workflows in the target repository.
    """
    github_token = get_setting("github.com", "token")
    api = GhApi(token=github_token)

    try:
        workflows = api.actions.list_repo_workflows(owner, repository)["workflows"]
    except HTTP404NotFoundError:
        return []

    workflows = [workflow for workflow in workflows if workflow["state"] == "active"]
    key_ids = [
        "node_id",
        "created_at",
        "updated_at",
        "url",
        "html_url",
        "badge_url",
        "state",
    ]
    for workflow in workflows:
        for del_key in key_ids:
            del workflow[del_key]
    return workflows



# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
"""Common helper functions."""

import re
from typing import Callable, Dict, Optional, Union

import gerrit_to_platform.github as github
from gerrit_to_platform.config import (
    Platform,
    ReplicationRemotes,
    get_replication_remotes,
)


def choose_dispatch(platform: Platform) -> Union[Callable, None]:
    """
    Choose platform job dispatcher.

    Args:
        platform (Platform): the platform that the dispatch is being looked up
            for

    Returns:
        Callable: The appropriate callable matching the dispatch_worfklow call
            signature for the platform
        None: If no dispatch_workflow is defined for the platform passed in a
            None is returned
    """
    if platform.value == "github":
        return github.dispatch_workflow

    return None


def choose_filter_workflows(platform: Platform) -> Union[Callable, None]:
    """
    Choose platform workflow filter.

    Args:
        platform (Platform): the platform that the filter_workflows is being
            looked up for


    Returns:
        Callable: The appropriate callable matching the filter_workflows call
            signature for the platform
        None: If no filter_workflows is defined for the platform passed in a
            None is returned
    """
    if platform.value == "github":
        return github.filter_workflows

    return None


def convert_repo_name(
    remotes: ReplicationRemotes, platform: Platform, remote: str, repository: str
) -> str:
    """
    Convert the repository name based on the remotenamestyle of the target
    platform/owner.

    Args:
        remotes (ReplicationRemotes): object containing the defined remotes
            styles
        platform (Platform): what platform is the conversion happening against
        remote (str): The specific remote that is being worked on
        repository (str): The repository name that needs conversion

    Returns:
        str: The repository name converted to the appropriate flattening for
            the target remote
    """

    remote_styles = {
        "dash": "-",
        "underscore": "_",
        "slash": "/",
    }

    target_style = remotes[platform.value][remote]["remotenamestyle"]
    converted_repository = repository.replace("/", remote_styles[target_style])

    return converted_repository


def find_and_dispatch(project: str, workflow_filter: str, inputs: Dict[str, str]):
    """
    Find relevant workflows and dispatch them.

    Args:
        project (str): the project repository name
        workflow_filter (str): the filter for the workflow names
        inputs (Dict[str, str]): key / value pair dictionary for inputs to be
            passed to the target workflow dispatch
    """
    remotes = get_replication_remotes()

    for platform in Platform:
        if platform.value not in remotes:
            continue

        dispatcher = choose_dispatch(platform)
        filter_workflows = choose_filter_workflows(platform)

        if dispatcher is None or filter_workflows is None:
            continue

        for remote in remotes[platform.value]:
            owner = remotes[platform.value][remote]["owner"]
            repo = convert_repo_name(remotes, platform, remote, project)
            workflows = filter_workflows(owner, repo, workflow_filter)

            for workflow in workflows:
                print(
                    f"Dispatching workflow '{workflow['name']}', "
                    + f"id {workflow['id']} on "
                    + f"{platform.value}:{owner}/{repo} for change "
                    + f"{inputs['GERRIT_CHANGE_NUMBER']} patch "
                    + inputs["GERRIT_PATCHSET_NUMBER"]
                )
                dispatcher(
                    owner,
                    repo,
                    workflow["id"],
                    f"refs/heads/{inputs['GERRIT_BRANCH']}",
                    inputs,
                )

            magic_repo = get_magic_repo(platform)
            if magic_repo:
                required_workflows = filter_workflows(
                    owner, magic_repo, workflow_filter, True
                )

                inputs["TARGET_REPO"] = f"{owner}/{repo}"

                for workflow in required_workflows:
                    print(
                        f"Dispatching required workflow '{workflow['name']}', "
                        + f"id {workflow['id']} on "
                        + f"{platform.value}:{owner}/{magic_repo} for change "
                        + f"{inputs['GERRIT_CHANGE_NUMBER']} patch "
                        + f"{inputs['GERRIT_PATCHSET_NUMBER']} against "
                        + f"{platform.value}:{owner}/{repo}"
                    )
                    dispatcher(
                        owner,
                        magic_repo,
                        workflow["id"],
                        "refs/heads/main",
                        inputs,
                    )


def get_change_id(change: str) -> str:
    """
    Get the Gerrit change_id from an hook event.

    Args:
        change (str): the change string passed to event handlers

    Returns:
        str: The extracted Gerrit change_id from the event hook
    """
    change_id_regex = r".*~.*~(I.*)"
    return re.findall(change_id_regex, change)[0]


def get_change_number(change_url: str) -> str:
    """
    Get the Gerrit change_number

    Args:
        change_url (str): the url to the specific change passed by the Gerrit
            event

    Returns:
        str: The change number as string extracted from the url
    """
    change_number_regex = r"^.*/\+/(\d+)$"
    return re.findall(change_number_regex, change_url)[0]


def get_change_refspec(change_number: str, patchset: str) -> str:
    """
    Return the change refspec from the change number (str) and patch number

    Args:
        change_number (str): The change number to work with
        patchset (str): The patchset number to work with

    Returns:
        str: The git refspec pointing to the hidden ref for the specific change
            and patchset
    """
    if int(change_number) < 100:
        ref_shard = change_number.zfill(2)
    else:
        ref_shard = change_number[len(change_number) - 2 :]
    return f"refs/changes/{ref_shard}/{change_number}/{patchset}"


def get_magic_repo(platform: Platform) -> Optional[str]:
    """
    Get the "magic" repo for a given Platform used to store organization wide
    required workflows

    Args:
        platform (Platform): Platform to lookup magic repo for

    Returns:
        Optional[str]: The magic repo name or None if not defined for the
            platform
    """

    if platform == Platform.GITHUB:
        return ".github"

    return None



# SPDX-License-Identifier: Apache-2.0
##############################################################################
# Copyright (c) 2023 The Linux Foundation and others.
#
# All rights reserved. This program and the accompanying materials are made
# available under the terms of the Apache-2.0 license which accompanies this
# distribution, and is available at
# https://opensource.org/licenses/Apache-2.0
##############################################################################
from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError
